from .impersonate.middleware import ImpersonateMiddleware
from .querystats.middleware import QueryStatsMiddleware


class StaffMiddleware:
    """All staff-related middleware in a single class."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return QueryStatsMiddleware(ImpersonateMiddleware(self.get_response))(request)



from typing import TYPE_CHECKING

from plain import models
from plain.models import Q
from plain.urls import reverse_lazy

from .base import (
    URL_NAMESPACE,
    StaffCreateView,
    StaffDeleteView,
    StaffDetailView,
    StaffListView,
    StaffUpdateView,
)

if TYPE_CHECKING:
    from plain import models
    from plain.views import View


def get_model_field(instance, field):
    if "__" in field:
        # Allow __ syntax like querysets use,
        # also automatically calling callables (like __date)
        result = instance
        for part in field.split("__"):
            result = getattr(result, part)

            # If we hit a None, just return it
            if not result:
                return result

            if callable(result):
                result = result()

        return result

    return getattr(instance, field)


class StaffModelListView(StaffListView):
    show_search = True
    allow_global_search = True

    model: "models.Model"

    fields: list = ["pk"]
    queryset_order = []
    search_fields: list = ["pk"]

    def get_title(self) -> str:
        if title := super().get_title():
            return title

        return self.model._meta.model_name.capitalize() + "s"

    @classmethod
    def get_nav_title(cls) -> str:
        if cls.nav_title:
            return cls.nav_title

        if cls.title:
            return cls.title

        return cls.model._meta.model_name.capitalize() + "s"

    @classmethod
    def get_slug(cls) -> str:
        return cls.model._meta.model_name

    def get_template_context(self):
        context = super().get_template_context()

        order_by = self.request.GET.get("order_by", "")
        if order_by.startswith("-"):
            order_by_field = order_by[1:]
            order_by_direction = "-"
        else:
            order_by_field = order_by
            order_by_direction = ""

        context["order_by_field"] = order_by_field
        context["order_by_direction"] = order_by_direction

        return context

    def get_objects(self):
        queryset = self.get_initial_queryset()
        queryset = self.order_queryset(queryset)
        queryset = self.search_queryset(queryset)
        return queryset

    def get_initial_queryset(self):
        # Separate override for the initial queryset
        # so that annotations can be added BEFORE order_by, etc.
        return self.model.objects.all()

    def order_queryset(self, queryset):
        if order_by := self.request.GET.get("order_by"):
            queryset = queryset.order_by(order_by)
        elif self.queryset_order:
            queryset = queryset.order_by(*self.queryset_order)

        return queryset

    def search_queryset(self, queryset):
        if search := self.request.GET.get("search"):
            filters = Q()
            for field in self.search_fields:
                filters |= Q(**{f"{field}__icontains": search})

            queryset = queryset.filter(filters)

        return queryset

    def get_field_value(self, obj, field: str):
        try:
            return super().get_field_value(obj, field)
        except AttributeError:
            return get_model_field(obj, field)

    def get_field_value_template(self, obj, field: str, value):
        templates = super().get_field_value_template(obj, field, value)
        if hasattr(obj, f"get_{field}_display"):
            # Insert before the last default template,
            # so it can still be overriden by the user
            templates.insert(-1, "staff/values/get_display.html")
        return templates


class StaffModelDetailView(StaffDetailView):
    model: "models.Model"
    fields: list = []

    def get_title(self) -> str:
        return str(self.object)

    @classmethod
    def get_slug(cls) -> str:
        return f"{cls.model._meta.model_name}_detail"

    @classmethod
    def get_path(cls) -> str:
        return f"{cls.model._meta.model_name}/<int:pk>/"

    def get_template_context(self):
        context = super().get_template_context()
        context["fields"] = self.fields or ["pk"] + [
            f.name for f in self.object._meta.get_fields() if not f.remote_field
        ]
        return context

    def get_field_value(self, obj, field: str):
        try:
            return super().get_field_value(obj, field)
        except AttributeError:
            return get_model_field(obj, field)

    def get_object(self):
        return self.model.objects.get(pk=self.url_kwargs["pk"])

    def get_template_names(self) -> list[str]:
        template_names = super().get_template_names()

        if not self.template_name and isinstance(self.object, models.Model):
            object_meta = self.object._meta
            template_names = [
                f"staff/{object_meta.package_label}/{object_meta.model_name}{self.template_name_suffix}.html"
            ] + template_names

        return template_names


class StaffModelCreateView(StaffCreateView):
    model: "models.Model"
    form_class = None  # TODO type annotation

    def get_title(self) -> str:
        if title := super().get_title():
            return title

        return f"Create {self.model._meta.model_name}"

    @classmethod
    def get_slug(cls) -> str:
        return f"{cls.model._meta.model_name}_create"

    @classmethod
    def get_path(cls) -> str:
        return f"{cls.model._meta.model_name}/create/"

    def get_template_names(self):
        template_names = super().get_template_names()

        if not self.template_name and isinstance(self.object, models.Model):
            object_meta = self.object._meta
            template_names = [
                f"staff/{object_meta.package_label}/{object_meta.model_name}{self.template_name_suffix}.html"
            ] + template_names

        return template_names


class StaffModelUpdateView(StaffUpdateView):
    model: "models.Model"
    form_class = None  # TODO type annotation
    success_url = "."  # Redirect back to the same update page by default

    def get_title(self) -> str:
        if title := super().get_title():
            return title

        return f"Update {self.object}"

    @classmethod
    def get_slug(cls) -> str:
        return f"{cls.model._meta.model_name}_update"

    @classmethod
    def get_path(cls) -> str:
        return f"{cls.model._meta.model_name}/<int:pk>/update/"

    def get_object(self):
        return self.model.objects.get(pk=self.url_kwargs["pk"])

    def get_template_names(self):
        template_names = super().get_template_names()

        if not self.template_name and isinstance(self.object, models.Model):
            object_meta = self.object._meta
            template_names = [
                f"staff/{object_meta.package_label}/{object_meta.model_name}{self.template_name_suffix}.html"
            ] + template_names

        return template_names


class StaffModelDeleteView(StaffDeleteView):
    model: "models.Model"

    def get_title(self) -> str:
        return f"Delete {self.object}"

    @classmethod
    def get_slug(cls) -> str:
        return f"{cls.model._meta.model_name}_delete"

    @classmethod
    def get_path(cls) -> str:
        return f"{cls.model._meta.model_name}/<int:pk>/delete/"

    def get_object(self):
        return self.model.objects.get(pk=self.url_kwargs["pk"])


class StaffModelViewset:
    @classmethod
    def get_views(cls) -> list["View"]:
        views = []

        if hasattr(cls, "ListView"):

            def get_list_url(self):
                return reverse_lazy(f"{URL_NAMESPACE}:{cls.ListView.view_name()}")

            for v in ["CreateView", "DetailView", "UpdateView", "DeleteView"]:
                if other_class := getattr(cls, v, None):
                    other_class.get_list_url = get_list_url
                    other_class.parent_view_class = cls.ListView

        if hasattr(cls, "CreateView"):

            def get_create_url(self):
                return reverse_lazy(f"{URL_NAMESPACE}:{cls.CreateView.view_name()}")

            if hasattr(cls, "ListView"):
                cls.ListView.get_create_url = get_create_url

        if hasattr(cls, "DetailView"):

            def get_detail_url(self, obj):
                return reverse_lazy(
                    f"{URL_NAMESPACE}:{cls.DetailView.view_name()}",
                    kwargs={"pk": obj.pk},
                )

            for v in ["ListView", "UpdateView", "DeleteView"]:
                if other_class := getattr(cls, v, None):
                    other_class.get_detail_url = get_detail_url

        if hasattr(cls, "UpdateView"):

            def get_update_url(self, obj):
                return reverse_lazy(
                    f"{URL_NAMESPACE}:{cls.UpdateView.view_name()}",
                    kwargs={"pk": obj.pk},
                )

            for v in ["ListView", "DetailView", "DeleteView"]:
                if other_class := getattr(cls, v, None):
                    other_class.get_update_url = get_update_url

        if hasattr(cls, "DeleteView"):

            def get_delete_url(self, obj):
                return reverse_lazy(
                    f"{URL_NAMESPACE}:{cls.DeleteView.view_name()}",
                    kwargs={"pk": obj.pk},
                )

            for v in ["ListView", "DetailView", "UpdateView"]:
                if other_class := getattr(cls, v, None):
                    other_class.get_delete_url = get_delete_url

        if hasattr(cls, "ListView"):
            views.append(cls.ListView)

        if hasattr(cls, "CreateView"):
            views.append(cls.CreateView)

        if hasattr(cls, "DetailView"):
            views.append(cls.DetailView)

        if hasattr(cls, "UpdateView"):
            views.append(cls.UpdateView)

        if hasattr(cls, "DeleteView"):
            views.append(cls.DeleteView)

        return views



from . import settings


def can_be_impersonator(user):
    return settings.IMPERSONATE_ALLOWED(user)


def can_impersonate_user(impersonator, target_user):
    if not can_be_impersonator(impersonator):
        return False

    # You can't impersonate staff users
    if target_user.is_staff:
        return False

    return True



import time
import traceback
from collections import Counter

import sqlparse

from plain.utils.functional import cached_property

IGNORE_STACK_FILES = [
    "threading",
    "socketserver",
    "wsgiref",
    "gunicorn",
    "whitenoise",
    "sentry_sdk",
    "querystats/core",
    "plain/template/base",
    "plain/utils/decorators",
    "plain/utils/deprecation",
    "plain/db",
    "plain/utils/functional",
    "plain/core/servers",
    "plain/core/handlers",
]


def pretty_print_sql(sql):
    return sqlparse.format(sql, reindent=True, keyword_case="upper")


def get_stack():
    return "".join(tidy_stack(traceback.format_stack()))


def tidy_stack(stack):
    lines = []

    skip_next = False

    for line in stack:
        if skip_next:
            skip_next = False
            continue

        if line.startswith('  File "') and any(
            ignore in line for ignore in IGNORE_STACK_FILES
        ):
            skip_next = True
            continue

        lines.append(line)

    return lines


class QueryStats:
    def __init__(self, include_tracebacks):
        self.queries = []
        self.include_tracebacks = include_tracebacks

    def __str__(self):
        s = f"{self.num_queries} queries in {self.total_time_display}"
        if self.duplicate_queries:
            s += f" ({self.num_duplicate_queries} duplicates)"
        return s

    def __call__(self, execute, sql, params, many, context):
        current_query = {"sql": sql, "params": params, "many": many}
        start = time.monotonic()

        result = execute(sql, params, many, context)

        if self.include_tracebacks:
            current_query["tb"] = get_stack()

        # if many, then X times is len(params)

        current_query["result"] = result

        current_query["duration"] = time.monotonic() - start

        self.queries.append(current_query)
        return result

    @cached_property
    def total_time(self):
        return sum(q["duration"] for q in self.queries)

    @staticmethod
    def get_time_display(seconds):
        if seconds < 0.01:
            return f"{seconds * 1000:.0f} ms"
        return f"{seconds:.2f} seconds"

    @cached_property
    def total_time_display(self):
        return self.get_time_display(self.total_time)

    @cached_property
    def num_queries(self):
        return len(self.queries)

    # @cached_property
    # def models(self):
    #     # parse table names from self.queries sql
    #     table_names = [x for x in [q['sql'].split(' ')[2] for q in self.queries] if x]
    #     models = connection.introspection.installed_models(table_names)
    #     return models

    @cached_property
    def duplicate_queries(self):
        sqls = [q["sql"] for q in self.queries]
        duplicates = {k: v for k, v in Counter(sqls).items() if v > 1}
        return duplicates

    @cached_property
    def num_duplicate_queries(self):
        # Count the number of "excess" queries by getting how many there
        # are minus the initial one (and potentially only one required)
        return sum(self.duplicate_queries.values()) - len(self.duplicate_queries)

    def as_summary_dict(self):
        return {
            "summary": str(self),
            "total_time": self.total_time,
            "num_queries": self.num_queries,
            "num_duplicate_queries": self.num_duplicate_queries,
        }

    def as_context_dict(self):
        # If we don't create a dict, the instance of this class
        # is lost before we can use it in the template
        for query in self.queries:
            # Add some useful display info
            query["duration_display"] = self.get_time_display(query["duration"])
            query["sql_display"] = pretty_print_sql(query["sql"])
            duplicates = self.duplicate_queries.get(query["sql"], 0)
            if duplicates:
                query["duplicate_count"] = duplicates

        summary = self.as_summary_dict()

        return {
            **summary,
            "total_time_display": self.total_time_display,
            "queries": self.queries,
        }

    def as_server_timing(self):
        duration = self.total_time * 1000  # put in ms
        duration = round(duration, 2)
        description = str(self)
        return f'querystats;dur={duration};desc="{description}"'



import plain.assets.urls
import plain.staff.urls
from plain.urls import include, path
from plain.views import View


class LoginView(View):
    def get(self):
        return "Login!"


class LogoutView(View):
    def get(self):
        return "Logout!"


urlpatterns = [
    path("staff/", include(plain.staff.urls)),
    path("assets/", include(plain.assets.urls)),
    path("login/", LoginView, name="login"),
    path("logout/", LogoutView, name="logout"),
]



from plain.urls import path

from .views import ImpersonateStartView, ImpersonateStopView

default_namespace = "impersonate"

urlpatterns = [
    path("stop/", ImpersonateStopView, name="stop"),
    path("start/<pk>/", ImpersonateStartView, name="start"),
]



from .base import Card


class TableCard(Card):
    template_name = "staff/cards/table.html"
    size = Card.Sizes.FULL

    headers = []
    rows = []
    footers = []

    def get_template_context(self):
        context = super().get_template_context()
        context["headers"] = self.get_headers()
        context["rows"] = self.get_rows()
        context["footers"] = self.get_footers()
        return context

    def get_headers(self):
        return self.headers.copy()

    def get_rows(self):
        return self.rows.copy()

    def get_footers(self):
        return self.footers.copy()



from enum import Enum

from plain.http import HttpRequest
from plain.templates import Template
from plain.utils.text import slugify
from plain.views import View


class Card:
    class Sizes(Enum):
        # Four column grid
        SMALL = 1
        MEDIUM = 2
        LARGE = 3
        FULL = 4

    template_name = "staff/cards/card.html"
    size: Sizes = Sizes.SMALL
    # unique_id: str  # Use for tying to dashboards, require it

    # Required fields
    title: str

    # Optional fields
    slug: str = ""
    description: str = ""
    text: str = ""
    link: str = ""
    number: int | None = None
    displays: list[str] | Enum | None = None

    # These will be accessible at render time
    view: View
    request: HttpRequest

    def render(self, view, request):
        self.view = view
        self.request = request
        return Template(self.template_name).render(self.get_template_context())

    @classmethod
    def view_name(cls) -> str:
        return f"card_{cls.get_slug()}"

    def get_template_context(self):
        context = {}

        context["size"] = self.size
        context["title"] = self.get_title()
        context["slug"] = self.get_slug()
        context["description"] = self.get_description()
        context["number"] = self.get_number()
        context["text"] = self.get_text()
        context["link"] = self.get_link()
        context["displays"] = self.get_displays()
        context["current_display"] = self.get_current_display()

        return context

    def get_title(self) -> str:
        return self.title

    def get_slug(self) -> str:
        return self.slug or slugify(self.title)

    def get_description(self) -> str:
        return self.description

    def get_number(self) -> int | None:
        return self.number

    def get_text(self) -> str:
        return self.text

    def get_link(self) -> str:
        return self.link

    def get_current_display(self) -> str:
        return self.request.GET.get(f"{self.get_slug()}.display", "")

    def get_displays(self) -> list[str] | Enum | None:
        if hasattr(self.displays, "copy"):
            # Avoid mutating the class attribute
            return self.displays.copy()
        else:
            return self.displays
